/* 
 * File:   gs_math.c
 * Author: eis_os (cim bytetransfer.de)
 *
 * Slow unoptimized math functions for quaternion
 * 
 * last update: 09.08.2011
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "gs_math.h"

const float DEGTORAD = 3.14159265359f / 180.0f;
const float RADTODEG = 180.0f / 3.14159265359f;

struct gs_vec vec_add(gs_vec l, gs_vec r) {
    gs_vec result;
    result.x = l.x + r.x;
    result.y = l.y + r.y;
    result.z = l.z + r.z;
    return result;
}

struct gs_vec vec_sub(gs_vec l, gs_vec r) {
    gs_vec result;
    result.x = l.x - r.x;
    result.y = l.y - r.y;
    result.z = l.z - r.z;
    return result;
}

struct gs_vec vec_crossproduct(gs_vec l, gs_vec r) {
    gs_vec result;
    result.x = l.y * r.z - l.z * r.y;
    result.y = l.z * r.x - l.x * r.z;
    result.z = l.x * r.y - l.y * r.x;
    return result;
}

float vec_dotproduct(gs_vec l, gs_vec r) {
    return l.x * r.x + l.y * r.y + l.z * r.z;
}


struct gs_vec vec_normalise(gs_vec vec) {
    float l = ( vec.x * vec.x + vec.y * vec.y + vec.z * vec.z);
    if ( l > 1e-08 ) {
        l = sqrtf(l);
        vec.x *= l;
        vec.x *= l;
        vec.x *= l;
    }
    return vec;
}

struct gs_vec vec_multiply_scalar(gs_vec l, float f) {
    gs_vec result;
    result.x = l.x * f;
    result.y = l.y * f;
    result.z = l.z * f;
    return result;
}

struct gs_quat quat_multiply(gs_quat l, gs_quat r) {
    gs_quat result;
    result.w = (r.w * l.w) - (r.x * l.x) - (r.y * l.y) - (r.z * l.z);
    result.x = (r.w * l.x) + (r.x * l.w) + (r.y * l.z) - (r.z * l.y);
    result.y = (r.w * l.y) + (r.y * l.w) + (r.z * l.x) - (r.x * l.z);
    result.z = (r.w * l.z) + (r.z * l.w) + (r.x * l.y) - (r.y * l.x);
    return result;
}
struct gs_quat quat_invert(gs_quat q) {
    gs_quat result;
    result.w = q.w;
    result.x = -q.x;
    result.y = -q.y;
    result.z = -q.z;
    return result;
}

struct gs_quat vec_to_quad(gs_vec vec) {
    gs_quat result;
    result.w = 0;
    result.x = vec.x;
    result.y = vec.y;
    result.z = vec.z;
    return result;
}
struct gs_vec quad_to_vec(gs_quat quat) {
    gs_vec result;
    result.x = quat.x;
    result.y = quat.y;
    result.z = quat.z;
    return result;
}

struct gs_vec quat_multiply_vec(gs_quat quat, gs_vec vec) {
	gs_quat vQuat, iQuat, result;
	
	vQuat = vec_to_quad(vec);
	iQuat = quat_invert(quat);
	result = quat_multiply(quat, quat_multiply(vQuat, iQuat));
	return quad_to_vec(result);
}


// Inspired by quaternion.h Irrlicht Engine
struct gs_vec quat_to_euler(gs_quat quat) {
	gs_vec result;
        double quatx2 = quat.x*quat.x;
        double quaty2 = quat.y*quat.y;
        double quatz2 = quat.z*quat.z;
        double quatw2 = quat.w*quat.w;
  
        result.z = (float) atan2(2.0*(quat.x*quat.y + quat.w*quat.z), quatw2 + quatx2 - quaty2 - quatz2);
        result.x = (float) atan2(2.0*(quat.y*quat.z + quat.w*quat.x), quatw2 - quatx2 - quaty2 + quatz2);
        result.y = asinf(-2.0f*(quat.x*quat.z - quat.w*quat.y));

        return result;
}

struct gs_quat euler_to_quat_alt(gs_vec vec) {
	gs_quat result;
        
        double sinx = sin(vec.x*0.5);
        double siny = sin(vec.y*0.5);
        double sinz = sin(vec.z*0.5);
      
        double cosx = cos(vec.x*0.5);
        double cosy = cos(vec.y*0.5);
        double cosz = cos(vec.z*0.5);
        
        result.w = (float) (cosx*cosy*cosz-sinx*siny*sinz);
        
        result.x = (float) (sinx*siny*cosz+cosx*cosy*sinz);
        result.y = (float) (sinx*cosy*cosz+cosx*siny*sinz);
        result.z = (float) (cosx*siny*cosz-sinx*cosy*sinz);
        
        return result;
}

struct gs_quat euler_to_quat(gs_vec vec) {
	gs_quat result;
        
        double sinx = sin(vec.x*0.5);
        double siny = sin(vec.y*0.5);
        double sinz = sin(vec.z*0.5);
      
        double cosx = cos(vec.x*0.5);
        double cosy = cos(vec.y*0.5);
        double cosz = cos(vec.z*0.5);
        
        result.w = (float) (cosx*cosy*cosz+sinx*siny*sinz);
        
        result.x = (float) (sinx*cosy*cosz-cosx*siny*sinz);
        result.y = (float) (cosx*siny*cosz+sinx*cosy*sinz);
        result.z = (float) (cosx*cosy*sinz-sinx*siny*cosz);
        
        return result;
}