/*
 * gs uncompress - unpacking files in gs format
 *
 * by eis_os (cim bytetransfer.de)
 * 
 * last update: 12.09.2012
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <zlib.h>

#include "gs.h"
#include "gs_filefunc.h"

#define TOOLNAME "gs_uncompress " GS_TOOLS_VERSION 

// options
static int force_overwrite = 0;


struct gs_file_entry_ll {
    struct gs_file_entry fe;
    struct gs_file_entry_ll *next;
    uint32_t filepos;
};



void usage(char *name) {
    printf("Usage: %s <options> file.gs\n", name);
}

static void chunk_pr_warn(struct gs_chunk *ch, const char *name, int level, int fd)
{
    print(level, "@ %#0.8x %s(%#08x) (%u bytes) \n", tell(fd)-8, name ? name : "UNKNOWN_CHUNK", ch->type, ch->size);
}


char* createoutputdirectory(char* filename) {
    int i = 0;
    char *directory = NULL;
    char *altdirectory = NULL;
    directory = filename_noext(filename);
    
    if (fileordir_present(directory)) {
        while(i < 100) {
            altdirectory = calloc(strlen(directory)+10, 1);
            sprintf(altdirectory, "%s.%u", directory, i);
            if (!fileordir_present(altdirectory)) {
                mkdir(altdirectory);
                free(directory);
                return altdirectory;
            }
            i++;
        }
        die("Output directory (%s) already exists , max tries reached",altdirectory);
    }
    mkdir(directory);
    return directory;
}

int mkdirforfile(char* filepath) {
    struct stat st;
    char *p, *mypath;
    mypath = strdup(filepath);
    p = mypath;
    
    while (*p == '/') p++;
    
    while (p = strchr (p, '/')) {
        *p = 0;
        if (stat(mypath, &st) == 0) {
            if (S_ISDIR(st.st_mode) == 0) {
                 die("can't create directory %s: file already exists\n", mypath);
            }
        } else {
            if (mkdir(mypath)) {
                die("can't create %s: %s\n", mypath, strerror(errno));
            }
        }
        *p++ = '/';
        while (*p == '/') p++;
    }
    if (stat(mypath, &st) == 0) {
        if (S_ISREG(st.st_mode)) {
            die("already exists %s\n", mypath);
        }
    }
    free(mypath);
    return 1;
}


int decompressfile(struct gs_file_entry_ll *ll, int fd) {
    unsigned char *inbuffer;
    unsigned char *outbuffer;
    int nb, ret;
    
    print(3, "%.128s\n", ll->fe.name);
    print(4, "insize: %6u, outsize: %6u, offset: %#08x; fileofs = %#08x\n", ll->fe.size_packed, ll->fe.size_unpacked, ll->fe.offset, ll->filepos);
    inbuffer = malloc(ll->fe.size_packed);
    if (!inbuffer) {
        die("inbuffer malloc failed\n");
    }
    outbuffer = malloc(ll->fe.size_unpacked);
    if (!outbuffer) {
        die("outbuffer malloc failed\n");
    }
    lseek(fd, ll->filepos, SEEK_SET);
    nb = read(fd, inbuffer, ll->fe.size_packed);
    if(nb < 0) {
        die("read(): %s\n", strerror(errno));
    } else if(nb == 0) {
        die("unexpected EOF\n");
    }
    uLongf written = ll->fe.size_unpacked;
    ret = uncompress(outbuffer, &written, inbuffer, ll->fe.size_packed);
    if (ret != Z_OK) {
        die("uncompressing error: %i\n", ret);
    }
    if (memcmp(outbuffer, &(ll->fe.signature), 4) != 0) {
        print(4, "-> signature broken! (compressed with GSFileViewer?)\n", ll->fe.name);
    }
    mkdirforfile(ll->fe.name);
    fd = open(ll->fe.name, CREAT_FLAGS, 0644);
    if(fd == -1)
        die("open output file failed: %s\n", strerror(errno));
    write(fd, outbuffer, ll->fe.size_unpacked);
    close(fd);
    
    free(inbuffer);
    free(outbuffer);
}


int extractgsarchive(char* filename) {
    int fd;
    unsigned int size;

    char *outputdirectory = NULL; 
    
    uint32_t num_files;
    struct gs_chunk ch_gsarchive, ch_gsfileindex, ch3;
    struct gs_file_entry fe;
    uint32_t gsarchive_offset;
    
    char *basepath = NULL;
    
    struct gs_file_entry_ll *ll = NULL;
    struct gs_file_entry_ll *llstart = NULL;
    struct gs_file_entry_ll *llprev = NULL;
    
    fd = gs_openfile(filename, FILETYPE_GS_ARCHIVE, "GS_ARCHIVE", &size);
    
    outputdirectory = createoutputdirectory(filename);
    print(0, "Created output directory %s\n", outputdirectory);
    
    if(chdir(outputdirectory) != 0)
        die("chdir('%s') failed: %s\n", outputdirectory, strerror(errno));
    
    while(size) {
        gs_read_chunk(fd, &ch_gsarchive, &size);
        if(ch_gsarchive.type != CHUNK_GS_ARCHIVE) {
            chunk_pr_warn(&ch_gsarchive,0, 1, fd);
            lseek(fd, ch_gsarchive.size, SEEK_CUR);
            continue;
        }
        gsarchive_offset = tell(fd);
        while(ch_gsarchive.size) {
            gs_read_chunk(fd, &ch_gsfileindex, &ch_gsarchive.size);

            switch(ch_gsfileindex.type) {
            case CHUNK_FILE_INDEX:
                gs_read(fd, &num_files, sizeof(num_files), &ch_gsfileindex.size);
                print(2, "NumberOfFiles: %u\n", num_files);

                while(ch_gsfileindex.size) {

                    gs_read_chunk(fd, &ch3, &ch_gsfileindex.size);

                            if(ch3.type == CHUNK_FILE_ENTRY) {
                                ll = calloc(sizeof(struct gs_file_entry_ll), 1);
                                gs_read(fd, &(ll->fe), sizeof(ll->fe), &ch3.size);
                                if (llprev == NULL) {
                                    llstart = ll;
                                } else {
                                    llprev->next = ll;
                                }
                                llprev = ll;
                                ll->filepos = gsarchive_offset + ll->fe.offset;
                                //print(3, "%6u (%6u)  %.128s  -> %#08x realofs = %#08x\n", ll->fe.size_packed, ll->fe.size_unpacked, ll->fe.name, ll->fe.offset, ll->filepos);
                            } else if (ch3.type == CHUNK_FILE_BASEPATH) {
                                basepath = calloc(ch3.size+1, 1);
                                gs_read(fd, basepath, ch3.size, &ch3.size);
                                print(3, "Warning: Base path found, folder structure needs fixup: %s\n", basepath);
                            } else if (ch3.type == CHUNK_FILE_FULLPATH) {
                                basepath = calloc(ch3.size+1, 1);
                                gs_read(fd, basepath, ch3.size, &ch3.size);
                                print(3, "Warning: Full path found, folder structure needs fixup: %s\n", basepath);    
                            } else {
                                chunk_pr_warn(&ch3, 0, 3, fd);
                    lseek(fd, ch3.size, SEEK_CUR);
                            }
                }
                break;
            case CHUNK_PACKED_DATA:
                //chunk_pr_warn(&ch_gsfileindex, "CHUNK_PACKED_DATA", 2, fd);
                // Simple skip data;
                lseek(fd, ch_gsfileindex.size, SEEK_CUR);
                break;
            default:
                chunk_pr_warn(&ch_gsfileindex, 0, 2, fd);
                lseek(fd, ch_gsfileindex.size, SEEK_CUR);
            }
        }
    }
    if (llstart) {
        ll = llstart;
        print(2, "Decompressing...\n");
        while(ll != NULL) {
                //print(3, "%6u (%6u)  %.128s  -> %#08x realofs = %#08x\n", ll->fe.size_packed, ll->fe.size_unpacked, ll->fe.name, ll->fe.offset, ll->filepos);
                decompressfile(ll, fd);
                ll = ll->next;
        }        
    }
    
    close(fd);
}


int main(int argc, char **argv)
{
    int i;
    char *filename = NULL, *directory = NULL, *path = NULL;
    char* basefolder = NULL;
    
    printf(TOOLNAME "\n");
    if(argc < 2) {
	usage(argv[0]);
	return -1;
    }
    
    for(i=1; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
            if (arg[1] == 'f') {
                force_overwrite = 1;
            } else if (arg[1] == 'b') {
                i++;
                if (i<argc) {
                        basefolder = strdup(argv[i]);
                } else {
                        usage(argv[0]);
                        return -1;
                }
            } else {
                printf("Unknown option %s \n", argv[i]);
                return -1;
            }
	} else {
            if(filename == NULL) {
                filename = strrchr(argv[i], PATH_DELIMITER);
                if(filename) {
                    *filename++ = '\0';
                    path = argv[i];
                    if(chdir(path) != 0)
                       die("chdir('%s') failed: %s\n", path, strerror(errno));
                } else {
                    path = ".";
                    filename = argv[i];
                }
            }
        }
    }
    if (!filename) {
        usage(argv[0]);
        return -1;
    }
    printf("working dir: %s\n", path);

    extractgsarchive(filename);
    return 0;
}
