/*
 * gs object export utility
 *
 * converts .object/.material => .obj/.mtl files
 *
 * by immi ( immi at gmx . org )
 * by eis_os (cim bytetransfer.de)
 * 
 * last update: 24.12.2011
 *
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "gs.h"
#include "gs_math.h"
#include "gs_filefunc.h"
#include "gs_path.h"

#define TOOLNAME "gs_export " GS_TOOLS_VERSION 

// options
int force_overwrite = 0;
int opt_dump_path = 1;


struct vertex_data {
    unsigned int num_vertices;
    struct gs_vertex verts[1];
};

struct index_data {
    unsigned int num_indices;
    uint32_t indices[1];
};

struct material {
    struct gs_material mat;
    const char *tex, *bump_map, *spec_map;
};

struct material_data {
    unsigned int num_materials;
    struct material material[1];
};

/* in gs_path.h 
struct path_data_ll {
    struct gs_path p;
    struct path_data_ll *next;
};
*/
struct obj_data {
    struct gs_object info;
    struct vertex_data *vdata;
    struct index_data *idata;
    struct material_data *mdata;
    struct path_data_ll *pdata;      
};


static struct vertex_data* load_vertex_data(int fd, unsigned int *size)
{
    uint32_t num;
    struct vertex_data *vdata;

    printf("loading vertex data ...\n");
    gs_read(fd, &num, sizeof(num), size);

    vdata = malloc(sizeof(*vdata) + (num-1) * sizeof(struct gs_vertex));
    vdata->num_vertices = num;

    gs_read(fd, vdata->verts, num * sizeof(struct gs_vertex), size);

    return vdata;
}


static struct index_data* load_index_data(int fd, unsigned int *size)
{
    uint32_t num;
    struct index_data *idata = 0;

    printf("loading indices ...\n");
    gs_read(fd, &num, sizeof(num), size);

    idata = malloc(sizeof(*idata) + (num-1) * sizeof(uint32_t));
    idata->num_indices = num;

    gs_read(fd, idata->indices, num * sizeof(uint32_t), size);

    return idata;
}


static struct material_data* load_material_data(int fd, unsigned int *size)
{
    uint32_t num, i = 0;
    struct gs_chunk ch;
    struct material_data *mdata;

    printf("loading material_data ...\n");
    gs_read(fd, &num, sizeof(num), size);

    mdata = malloc(sizeof(*mdata) + (num-1) * sizeof(struct material));
    memset(mdata, 0, sizeof(*mdata) + (num-1) * sizeof(struct material));
    mdata->num_materials = num;

    while(*size) {
	gs_read_chunk(fd, &ch, size);

	if(ch.type == CHUNK_MATERIAL_DEF) {
	    if(i == num) // safety check
		die("contains %u MATERIAL_DEFs, expected only %u\n", i+1, num);

	    gs_read(fd, &mdata->material[i].mat, sizeof(struct gs_material),
		       &ch.size);
	    i++;
	}

	lseek(fd, ch.size, SEEK_CUR); // skip not consumed data
    }

    if(i != num)
	die("only %u materials defined - expected %u\n",i, num);

    return mdata;
}


static void load_obj_data(int fd, struct obj_data* obj, unsigned int *size)
{
    while(*size) {
	struct gs_chunk ch;
	gs_read_chunk(fd, &ch, size);

	if(ch.type == CHUNK_VERTEX_DATA)
	    obj->vdata = load_vertex_data(fd, &ch.size);
	else if(ch.type == CHUNK_INDEX_DATA)
	    obj->idata = load_index_data(fd, &ch.size);
	else if(ch.type == CHUNK_MATERIAL_DATA)
	    obj->mdata = load_material_data(fd, &ch.size);

	lseek(fd, ch.size, SEEK_CUR); // skip not consumed data
    }
}

static void load_obj_path_data(int fd, struct obj_data* obj, unsigned int *size)
{
    uint32_t num;
    struct path_data_ll *data = 0;
    struct path_data_ll **pointer = &obj->pdata;

    //printf("Loading path");
    
    while (*pointer != 0) {
        pointer = &(*pointer)->next;
    }
    
    data = malloc(sizeof(struct path_data_ll));
    gs_read(fd, &data->p, sizeof(data->p), size);
    data->next = NULL;
   // printf(" '%.32s' ", data->p.name);
    if (data->p.flags != 0) die("Path Error: '%.32s' flags needs to be zero, but '%#08x' found", data->p.name, data->p.flags);
    *pointer = data;
}

static struct obj_data* load_object(int fd, unsigned int size)
{

    struct obj_data *obj = malloc(sizeof(*obj));
    obj->vdata = NULL;
    obj->idata = NULL;
    obj->mdata = NULL;
    obj->pdata = NULL;

    printf("loading object:");
    gs_read(fd, &obj->info, sizeof(struct gs_object), &size);
    printf(" '%s'\n", obj->info.name);

    while(size) {
	struct gs_chunk data;
	gs_read_chunk(fd, &data, &size);

	if(data.type == CHUNK_OBJECT_DATA)
	    load_obj_data(fd, obj, &data.size);
        if(opt_dump_path && data.type == CHUNK_OBJ_PATH_INFO)
	    load_obj_path_data(fd, obj, &data.size);

	lseek(fd, data.size, SEEK_CUR); // not interested. skip
    }

    // safety checks
    if(!obj->vdata || !obj->idata || !obj->mdata) return NULL;
    if(obj->idata->num_indices % 3  != 0)
	die("num_indices %% 3 != 0: no triangles ??\n");

    return obj;
}




static const char* read_tex_ref(int fd, unsigned int *size, const char *path)
{
    struct gs_tex tex;
    char *tex_file;
    gs_read(fd, &tex, sizeof(tex), size);

    tex_file = make_path(NULL, tex.file, 0);

    if(!file_exists(tex_file))
	die("referenced texture not found: %s\n", tex_file);

    return tex_file;
}

static void load_material_file(int fd, unsigned int size, struct material *m,
			      const char *path)
{
    struct gs_chunk ch;
    struct gs_material mat;

    gs_read(fd, &mat, sizeof(mat), &size);

    while(size) {
	gs_read_chunk(fd, &ch, &size);

	if(ch.type == CHUNK_MAT_TEXTURE)
	    m->tex = read_tex_ref(fd, &ch.size, path);
	else if(ch.type == CHUNK_MAT_TEXTURE2)
	    m->bump_map = read_tex_ref(fd, &ch.size, path);
	else if(ch.type == CHUNK_MAT_TEXTURE3)
	    m->spec_map = read_tex_ref(fd, &ch.size, path);

	lseek(fd, ch.size, SEEK_CUR); // skip any data not consumed
    }
}


static void write_mtl_file(const char *mtl_file, struct obj_data *obj)
{
    unsigned int i;
    struct material *m = obj->mdata->material;
    FILE *f;

    printf("writing mtl file: %s\n", mtl_file);
    if(!force_overwrite && file_exists(mtl_file))
        die("'%s' already exists. Remove or use -f to force overwrite\n",
            mtl_file);

    f = fopen(mtl_file, "w");
    if(!f)
	die("fopen(%s): %s\n", mtl_file, strerror(errno));

    fprintf(f, "# mtl created by " TOOLNAME "\n\n");

    for(i = 0; i < obj->mdata->num_materials; i++, m++) {
	fprintf(f, "\nnewmtl %s\n", m->mat.name);
        
        fprintf(f, "d %f\n", (float) m->mat.opacity / 255.0f);  
        fprintf(f, "Ka %f %f %f\n", (float) m->mat.ambient.r / 255.0f, (float)m->mat.ambient.g / 255.0f, (float)m->mat.ambient.b / 255.0f);  
      	fprintf(f, "Kd %f %f %f\n", (float) m->mat.diffuse.r / 255.0f, (float)m->mat.diffuse.g / 255.0f, (float)m->mat.diffuse.b / 255.0f);  
        fprintf(f, "Ks %f %f %f\n", (float) m->mat.specular.r / 255.0f, (float)m->mat.specular.g / 255.0f, (float)m->mat.specular.b / 255.0f);  
        
        if(m->tex)
	    fprintf(f, "map_Kd %s\n", m->tex);
	if(m->spec_map)
	    fprintf(f, "map_Ks %s\n", m->spec_map);
	if(m->bump_map)
	    fprintf(f, "map_bump %s\n", m->bump_map);
    }

    fclose(f);
}

static void write_obj_file(const char *obj_file, struct obj_data *obj,
			   const char *mtl_file)
{
    unsigned int i,j;
    FILE *f;

    printf("writing obj file: %s\n", obj_file);
    if(!force_overwrite && file_exists(obj_file))
        die("'%s' already exists. Remove or use -f to force overwrite\n",
            obj_file);

    f = fopen(obj_file, "w");
    if(!f)
	die("fopen(%s): %s\n", obj_file, strerror(errno));

    fprintf(f, "# obj created by " TOOLNAME "\n\n");

    fprintf(f, "mtllib %s\n", mtl_file);
    fprintf(f, "o %s\n", obj->info.name);

    // write out vertex data
    // note: as trial and error has shown the v-coordinate has to be
    // flipped to work correctly in blender
    struct gs_vertex *v = obj->vdata->verts;
    fprintf(f, "# num vertices: %u", obj->vdata->num_vertices);
    for(i = 0; i < obj->vdata->num_vertices; i++,v++) {
        fprintf(f, "\nv %g  %g  %g\n", v->pos.x, v->pos.y, v->pos.z);
	fprintf(f, "vt %g %g\n", v->uv[0], -v->uv[1]);
	fprintf(f, "vn %g %g %g\n", v->nor.x, v->nor.y, v->nor.z);
    }

    fprintf(f, "\n# indices: %u\n", obj->idata->num_indices);
    struct material *m = obj->mdata->material;
    uint32_t *idx = obj->idata->indices;
    for(i = 0; i < obj->mdata->num_materials; i++, m++) {
	fprintf(f, "\nusemtl %s\n\n", m->mat.name);

	idx = obj->idata->indices + m->mat.idx_start;
        
        if (m->mat.idx_num > obj->idata->num_indices) {
              die("Material '%s' index count '%u' outside max indices '%u'\n", m->mat.name, m->mat.idx_num, obj->idata->num_indices);
        }
        if (m->mat.idx_start > obj->idata->num_indices) {
         //   die("Material index start '%u' for material '%s' outside max indices '%u'", m->mat.idx_num, m->mat.idx_start,  m->mat.name,  obj->idata->num_indices);
              die("Material '%s' index start '%u' count '%u' outside max indices '%u'\n", m->mat.name, m->mat.idx_start,  m->mat.idx_num,  obj->idata->num_indices);
        }
        print(1, "Material '%s' index start '%u' count '%u' / max indices '%u'\n", m->mat.name, m->mat.idx_start,  m->mat.idx_num,  obj->idata->num_indices);
	for(j = 0; j < m->mat.idx_num; j++, idx++) {
	    if( j % 3 == 0) fprintf(f, "\nf");
	     // note: obj file uses 1-based array indices
	    fprintf(f, " %u/%u/%u", *idx + 1, *idx + 1, *idx + 1);
	}
	fprintf(f, "\n");
    }

    fclose(f);
}


static void write_path_file(const char *obj_path_file, struct obj_data *obj) {
    struct path_data_ll *pd;
    unsigned int i,j;
    FILE *f;

    if (obj->pdata == NULL) return;
    pd = obj->pdata;
    
    printf("writing path file: %s\n", obj_path_file);
    if(!force_overwrite && file_exists(obj_path_file))
        die("'%s' already exists. Remove or use -f to force overwrite\n",
            obj_path_file);

    f = fopen(obj_path_file, "w");
    if(!f)
	die("fopen(%s): %s\n", obj_path_file, strerror(errno));
    
    fprintf(f, "# path file created by " TOOLNAME "\n");
    fprintf(f, "version 1\n\n");
   
    while(pd > 0) { 
        gs_vec position = pd->p.pos;
        gs_quat qrotation = pd->p.quat;

        fprintf(f, "name %s\n", pd->p.name);
        fprintf(f, "pos %g %g %g\n", position.x, position.y, position.z);
        fprintf(f, "quat %g %g %g %g\n", qrotation.x, qrotation.y, qrotation.z, qrotation.w);
        
        gs_vec degrot;
        gs_vec erot = quat_to_euler(qrotation);
        degrot.x = erot.x * RADTODEG;
        degrot.y = erot.y * RADTODEG;
        degrot.z = erot.z * RADTODEG;
        fprintf(f, "#rotation %g %g %g\n\n", degrot.x, degrot.y, degrot.z);
        
       /* 
        gs_vec radrot;
        radrot.x = degrot.x * DEGTORAD;
        radrot.y = degrot.y * DEGTORAD;
        radrot.z = degrot.z * DEGTORAD;
        gs_quat testquat = euler_to_quat(radrot);
       
        fprintf(f, "testquat %g %g %g %g\n", testquat.x, testquat.y, testquat.z, testquat.w);
        fprintf(f, "\n");
       */
        
        pd = pd->next;
    }
    fclose(f);
}


// TODO: replace with common gs_openfile
// type is either FILETYPE_GS_MATERIAL || FILE_TYPE_GS_OBJECT
// size contains the size of data in the file
// returns filedescriptor to open file or -1 on error
static int open_gs_file(const char *filename, uint32_t type,
			unsigned int *size)
{
    struct gs_chunk file;
    int fd;

    printf("reading: '%s'\n", filename);
    fd = open(filename, OPEN_FLAGS);
    if(fd == -1)
	die("open(): %s\n", strerror(errno));

    read(fd, &file, sizeof(file));
    if(file.type != type)
	die("%s: not an %s file\n", filename,
	    (type == FILETYPE_GS_OBJECT) ? "object" : "material");

    *size = file.size;
    return fd;
}


static int load_file(char *filename)
{
    const char *path = ".";
    unsigned int size;
    int fd;

    char *file = strrchr(filename, PATH_DELIMITER);
    if(file) {
	*file = '\0';
	file++;
	path = filename;
	if(chdir(path) != 0)
	    die("chdir('%s') failed: %s\n", path, strerror(errno));
    } else
	file = filename;

    printf("working dir: '%s'\n", path);

    fd = open_gs_file(file, FILETYPE_GS_OBJECT, &size);

    struct obj_data *obj = load_object(fd, size);
    close(fd);

    if(!obj) return -1;

    // load material files
    printf("looking for material files\n");
    unsigned int i;
    struct material *mat = obj->mdata->material;
    for(i = 0; i < obj->mdata->num_materials; i++, mat++) {
	char *mfile = make_path(NULL, mat->mat.name, ".material");

	fd = open_gs_file(mfile, FILETYPE_GS_MATERIAL, &size);
	free(mfile);

	load_material_file(fd, size, mat, path);

	close(fd);
    }

    // now write output - create file in current directory
    char *mtl_file = make_path(0, obj->info.name, ".mtl");
    char *obj_file = make_path(0, obj->info.name, ".obj");
    char *path_file = make_path(0, obj->info.name, ".paths");
    char *path_obj_file = make_path(0, obj->info.name, "_paths.obj");
     
    write_mtl_file(mtl_file, obj);
    write_obj_file(obj_file, obj, mtl_file);
    write_path_file(path_file, obj);
    //write_obj_path_file(path_obj_file, obj);
    write_obj_path_file_combined(path_obj_file, force_overwrite, obj->pdata, TOOLNAME);

    
    printf("export successfull\n");
    return 0;
}

void usage(char* name) {
    printf("Usage: %s <options> <file>\n"
    " -f force overwrite\n"
    " -p export object paths (unused)\n", name);
}

int main(int argc, char **argv)
{
    int i;
    char *filename = NULL;
    
    printf(TOOLNAME "\n");
    if(argc < 2) {
        usage(argv[0]);
	return -1;
    }
    
    for(i=1; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
            if (arg[1] == 'f') {
                force_overwrite = 1;
            } else if (arg[1] == 'p') {
                opt_dump_path = 1;
            } else {
                printf("Unknown option %s \n", argv[i]);
                usage(argv[0]);
                return -1;
            }
	} else {
            filename = argv[i];
        }
    }
    if (filename) {
        load_file(filename);
        
        return 0;
    }
    usage(argv[0]);
    return -1;
}
