/*
 * gs object import utility
 *
 * converts .obj/.mtl  =>  .object/.material files
 *
 * note: this is very much work in progress
 *
 * by immi ( immi at gmx . org )
 * by eis_os (cim bytetransfer.de)
 * 
 * last update: 27.05.2012
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "gs.h"
#include "gs_math.h"
#include "gs_path.h"
#include "gs_filefunc.h"

#define TOOLNAME "gs_import " GS_TOOLS_VERSION 

// 32k vertices should really be enough
#define MAX_LIST_SIZE 32 * 1024

static int force_overwrite;
static int verbose;

static char *obj_name;

// place to store v/vn/vt entries
static unsigned int num_v, num_vn, num_vt;
static float *v_list, *vn_list, *vt_list;

// while the OBJ file format has separate indices for vertices/tex coordinate
// & normals - CiM uses just one index for each v/vt/vn triplet
// keep these triplets in an ordered list so we can avoid adding duplicates
unsigned int nventries;
struct v_entry {
    unsigned int idx;
    unsigned int v_idx, vt_idx, vn_idx;
    struct v_entry *next;
};

struct v_entry *v_entries; // list head

unsigned int nindices;
struct v_entry **v_indices;


struct material {
    char *name;
    char *texture;
    uint8_t material_used; 
    struct gs_material* gsmat;
    struct gs_tex* diffusemap;
    struct gs_tex* normalmap;
    struct gs_tex* specularmap;
    /* struct gs_tex* illumap;
    struct gs_tex* aomap; */
    
//    unsigned int start_idx, num_idx;
    struct material *next;
};

unsigned int nmaterials;
struct material *material_list;
struct material *act_mat;

struct path_data_ll *path_list;


static void read_floats_to_list(char *p, float *array, unsigned int elem_count,
			unsigned int num, struct parser* parser)
{
    unsigned int i;
    float *dst = array + num * elem_count;

    if(elem_count == MAX_LIST_SIZE)
	die("MAX_LIST_SIZE(%u) reached (line: %u)\n", MAX_LIST_SIZE, parser->nline);

    for(i = 0; i < num; i++)
	dst[i] = read_float(p, &p, parser);
}


// check if idx triple already exists otherwise create a new entry
// TODO: rework so you can unterstand the logic with getting a headache :o
static struct v_entry *get_v_entry(unsigned int v_idx, unsigned int vt_idx,
				   unsigned int vn_idx)
{
    struct v_entry *p, *e = NULL, *last = NULL;
    struct v_entry *res = NULL;

    p = v_entries;
    while(1) {
	last = e; // keep ptr to last elemnt
	if(!p) break;
	e = p;
	p = p->next;
	if(e->v_idx < v_idx) continue;
	else if(e->v_idx > v_idx) break; // enry doesn't exist

	// v_idx == e->v_idx
	if(e->vt_idx < vt_idx) continue;
	else if(e->vt_idx > vt_idx) break;

	// vt_idx == e->vt_idx
	if(e->vn_idx  < vn_idx) continue;
	else if(e->vn_idx > vn_idx) break;

	// found
	res = e;
	break;
    }
    if(res)
	return res;

    // create element;
    res = malloc(sizeof(struct v_entry));
    res->v_idx = v_idx;
    res->vt_idx = vt_idx;
    res->vn_idx = vn_idx;
    res->next = NULL;
    nventries++;
    //printf("insert %u/%u/%u\n", v_idx, vt_idx, vn_idx);

    if(last) {
	res->next = last->next;
	last->next = res;
    } else {
	res->next = v_entries;
	v_entries = res;
    }

    return res;
}

// parse index triplet: v/vt/vn
// note: OBJ file uses 1-base array indices -> converting
// to zero-base indices
static struct v_entry* parse_indices(char *p, char **end, struct parser* parser)
{
    unsigned int v_idx, vt_idx, vn_idx;
    parser->errinfo = "face definition needs vertex & tex & normal index";
    v_idx = read_uint(p, &p, parser) - 1;
    if(*p == '/') {
        parser->errinfo = "missing texture index, uv map coordinates created and exported?";
	vt_idx = read_uint(p+1, &p, parser) - 1;
	if(*p == '/') {
            parser->errinfo = "missing normal index, exported normals?";
	    vn_idx = read_uint(p+1, &p, parser) - 1;
	    *end = p;
	    return get_v_entry(v_idx, vt_idx, vn_idx);
	}
    }
    die_parser(parser, p, "expected index or /");
    //die("face definition needs vertex & tex & normal index on line %u\n", parser->nline);
}

// assume triangle
static void read_face_def(char *p, struct parser* parser)
{
    if(nindices == MAX_LIST_SIZE)
	die("reached MAX_LIST_SIZE(%u)\n", MAX_LIST_SIZE);

    struct v_entry **v = v_indices + 3 *nindices;
    v[0] = parse_indices(p, &p, parser);
    v[1] = parse_indices(p, &p, parser);
    v[2] = parse_indices(p, &p, parser);
    nindices++;
}

struct material* material_exists(struct material *m, const char* materialname)
{
    if(m) {
        do {
           if (strcmp(m->name, materialname) == 0) {
               return m;
           } 
           m = m->next;
        } while(m);
    }
    return NULL;
}

static void read_mtl_file(const char *filename)
{
    struct parser *parser;
    struct material *m = NULL; // current material
    struct gs_tex *tex;
    
    parser = init_parser(filename);
    while(parser->line) {
        char *line = parser->line;
        if( strncmp(line, "newmtl ", 7) == 0) {
            if(verbose) {
                print(1, "newmtl: %s @%u\n", line+7, parser->nline); 
            }
            
            m = material_exists(material_list, line+7);
            if (m) {
                print(0, "[warn] Material %s already defined, old material data will be overwritten\n", m->name);
            } else {   
                struct material *newm = calloc(1, sizeof(*newm));
                struct gs_material *gsmat = calloc(1, sizeof(*gsmat));

                newm->name = strdup(line+7);
                if(material_list != 0) {
                    m = material_list;
                    while(m->next != 0) {
                        m = m->next;
                    }
                    m->next = newm;
                } else {
                    material_list = newm;
                }
                m = newm;

                strncpy(gsmat->name, m->name, 32);
                gsmat->type = 0;
                gsmat->flags = 0;
                gsmat->twoSided = 0;
                gsmat->opacity = 0xFF;
                //mat.unk[1] = 0x000400;
                gsmat->specularLevel = 0x00;
                gsmat->glossiness = 0x04;
                gsmat->brightness = 0;
                gsmat->reflection = 0;

                gsmat->ambient.abgr = 0xff808080;
                gsmat->diffuse.abgr = 0xff808080;
                gsmat->specular.abgr = 0xffffffff;
                gsmat->illumination.abgr = 0xffffffff;

                gsmat->idx_start = 0;
                gsmat->idx_num = 0;
                gsmat->vert_start = 0;
                gsmat->vert_num = 0;
                m->gsmat = gsmat;

                nmaterials++;
            }
            
        } else if(strnicmp(line, "d ", 2) == 0) {
            if(!m)
		die("missing newmtl statement\n");
            char *p = line+2;
            float fopacity = read_float(p, &p, parser);
            m->gsmat->opacity = (fopacity * 255.0f);        
            if(verbose) {
                print(2, "d: opacity: 0x%02x @%u\n",  m->gsmat->opacity, parser->nline); 
            }
         } else if(strnicmp(line, "Ka ", 3) == 0) {
            if(verbose) {
                print(2, "Ka: unsupported currently @%u\n", parser->nline); 
            }
         } else if(strnicmp(line, "gsflags ", 8) == 0) {
            if(!m)
                die("missing newmtl statement\n");
            char *p = line+8;
            unsigned int flags = read_uint(p, &p, parser);
            m->gsmat->flags = flags;
            if(verbose) {
                print(2, "gsflags: 0x%02x @%u\n",  m->gsmat->flags, parser->nline); 
            }
         } else if(strnicmp(line, "map_Kd_alpha ", 13) == 0) {
            char *p = line+13;
            unsigned int alphaChannel = read_uint(p, &p, parser);
            unsigned int alphaBits = read_uint(p, &p, parser);
            if (!m->diffusemap) {
                die("map_Kd not defined yet on line %u\n", parser->nline);
            }
            m->diffusemap->alphaChannel = alphaChannel;
            m->diffusemap->alphaBits = alphaBits;
    
            if(verbose) {
                print(2, "map_Kd_alpha: alphaChannel: 0x%02x alphaBits: 0x%02x @%u\n",   m->diffusemap->alphaChannel, m->diffusemap->alphaBits, parser->nline); 
            }
         } else if(strnicmp(line, "map_Ks_alpha ", 13) == 0) {
            char *p = line+13;
            unsigned int alphaChannel = read_uint(p, &p, parser);
            unsigned int alphaBits = read_uint(p, &p, parser);
            if (!m->specularmap) {
                die("map_Ks not defined yet on line %u\n", parser->nline);
            }
            m->specularmap->alphaChannel = alphaChannel;
            m->specularmap->alphaBits = alphaBits;
    
            if(verbose) {
                print(2, "map_Ks_alpha: alphaChannel: 0x%02x alphaBits: 0x%02x @%u\n",   m->specularmap->alphaChannel, m->specularmap->alphaBits, parser->nline); 
            }
         } else if(strnicmp(line, "map_bump_alpha ", 15) == 0) {
            char *p = line+15;
            unsigned int alphaChannel = read_uint(p, &p, parser);
            unsigned int alphaBits = read_uint(p, &p, parser);
            if (!m->normalmap) {
                die("map_bump not defined yet on line %u\n", parser->nline);
            }
            m->normalmap->alphaChannel = alphaChannel;
            m->normalmap->alphaBits = alphaBits;
    
            if(verbose) {
                print(2, "map_bump_alpha: alphaChannel: 0x%02x alphaBits: 0x%02x @%u\n", m->normalmap->alphaChannel, m->normalmap->alphaBits, parser->nline); 
            }
         } else if(strnicmp(line, "map_Kd ", 7) == 0) {
	    if(!m)
		die("missing newmtl statement\n");
            m->texture = strdup(line + 7);
            if(!file_exists(line+7))
                die("%s: referenced texture doesn't exist '%s' on line %u\n",
		    filename, line+7, parser->nline);    
            if(verbose) {
                print(2, "map_Kd:   texture: '%s' @%u\n", line+7, parser->nline); 
            }
            tex = calloc(1, sizeof(struct gs_tex));
            strncpy(tex->file, line+7, 32);
            tex->type = 0x00;
            tex->amount = 0xFF;
            tex->alphaChannel = 0;
            tex->alphaBits = 8; // stations have 8?;
            tex->tileU = 0xFF;
            tex->tileV = 0xFF;
            tex->mirrorU = 0;
            tex->mirrorV = 0;
            m->diffusemap = tex;
            
        } else if(strnicmp(line, "map_Ks ", 7) == 0) {
            if(!m)
            die("missing newmtl statement on line %u\n", parser->nline);

            if(!file_exists(line+7))
            die("%s: referenced texture doesn't exist: '%s' on line %u\n",
		    filename, line+7, parser->nline);    
            if(verbose) {
                print(2, "map_Ks:   texture: '%s' @%u\n", line+7, parser->nline); 
            }
            tex = calloc(1, sizeof(struct gs_tex));
            strncpy(tex->file, line+7, 32);
            
            tex->type = 0x00;
            tex->amount = 0xFF;
            tex->alphaChannel = 0;
            tex->alphaBits = 0;
            tex->tileU = 0xFF;
            tex->tileV = 0xFF;
            tex->mirrorU = 0;
            tex->mirrorV = 0;
            m->specularmap = tex;
	} else if(strnicmp(line, "map_bump ", 9) == 0) {
            if(!m)
            die("missing newmtl statement\n");

            if(!file_exists(line+9))
            die("%s: referenced texture doesn't exist: '%s' on line %u\n",
		    filename, line+9, parser->nline);    
            if(verbose) {
                print(2, "map_bump: texture: '%s' @%u\n", line+9, parser->nline); 
            }
            tex = calloc(1, sizeof(struct gs_tex));
            strncpy(tex->file, line+9, 32);
            
            tex->type = 0x00;
            tex->amount = 0xFF;
            tex->alphaChannel = 1;
            tex->alphaBits = 8;
            tex->tileU = 0xFF;
            tex->tileV = 0xFF;
            tex->mirrorU = 0;
            tex->mirrorV = 0;
            m->normalmap = tex;
	}

	parser_read_line(parser); // read next line
    }
    close(parser->fd);
    free(parser);
}


static void set_active_material(const char* name, unsigned int linenumber)
{
    struct material *m = material_list;
    while(m) {
	if(strcmp(m->name, name) == 0)
	    break;
	m = m->next;
    }
    if(!m)
	die("'%s' - unknown material on line %u\n", name, linenumber);
    
    // Material will be the same, no need to change
    if (act_mat == m)
        return;
    
    if (m->material_used > 0)
        die("'%s' material reusing not supported yet on line %u\n", name, linenumber);
    
    m->material_used = 1;
    m->gsmat->idx_start = nindices * 3;
    m->gsmat->vert_start = nventries;

//  m->start_idx = nindices * 3;
    if(act_mat) {
//      act_mat->gsmat->idx_num = (nindices - act_mat->gsmat->idx_start)*3;
        act_mat->gsmat->idx_num = nindices*3 - act_mat->gsmat->idx_start;
        act_mat->gsmat->vert_num = nventries - act_mat->gsmat->vert_start;
// 	act_mat->num_idx = (nindices - act_mat->start_idx)*3;
    }
    act_mat = m;
}


static void read_obj_parse_line(struct parser* parser)
{
    char *p = parser->line;

    switch(*p++) {
    case 'v':
	if(*p == ' ') {
	    read_floats_to_list(p + 1, v_list, num_v++, 3, parser);
	} else if( *(p+1) == ' ') {
	    switch(*p) {
	    case 't': read_floats_to_list(p+2, vt_list, num_vt++, 2, parser); break;
	    case 'n': read_floats_to_list(p+2, vn_list, num_vn++, 3, parser); break;
	    default: break;
	    }
	}
	break;
    case 'f':
	if(*p == ' ')
	    read_face_def(p + 1, parser);
	break;
    case 'o':
	if (*p == ' ') { // o <obj name>
	    if(obj_name) // already have a named object
		die("can't handle more than one obj mesh in file on line %u\n", parser->nline);
            if(strncmp(p+1, "Mesh_", 5) == 0) {
                obj_name = strdup(p + 1 + 5);
            } else {
                obj_name = strdup(p + 1);
            }
	}
	break;
    case 'm':
	if(strncmp(p, "tllib ", 6) == 0)
	    read_mtl_file(p + 6);
	break;
    case 'u':
	if(strncmp(p, "semtl ", 6) == 0) // set active material
	    set_active_material(p + 6, parser->nline);
	break;
    default: break;
    }
}


static int read_obj_file(const char *filename)
{
    struct parser *parser = init_parser(filename);
    while(parser->line) {
	read_obj_parse_line(parser);
	parser_read_line(parser); // read next line
    }

    if(act_mat) {
//      act_mat->gsmat->idx_num = (nindices - act_mat->gsmat->idx_start)*3;
        act_mat->gsmat->idx_num = nindices*3 - act_mat->gsmat->idx_start;
        act_mat->gsmat->vert_num = nventries - act_mat->gsmat->vert_start;
//      act_mat->num_idx = (nindices - act_mat->start_idx)*3;
    }
    return 0;
}

static void write_object_file(const char *filename)
{
    printf("writing: '%s'\n", filename);

    if(!force_overwrite && file_exists(filename))
	die("'%s' already exists. Remove or use -f to force overwrite\n",
	    filename);

    //calc sizes
    unsigned int vdata_len = sizeof(struct gs_vertex) * nventries
			     + sizeof(uint32_t);
    unsigned int idata_len = sizeof(uint32_t) * 3 * nindices
			     + sizeof(uint32_t);
    unsigned int mdata_len = (sizeof(struct gs_material)+sizeof(struct gs_chunk)) * nmaterials + sizeof(uint32_t);

    unsigned int objdata_len = vdata_len + idata_len + mdata_len
			      + sizeof(struct gs_chunk) * 3;

    unsigned int obj_len = objdata_len + sizeof(struct gs_chunk) +
			   sizeof(struct gs_object);

    int fd = open(filename, CREAT_FLAGS, 0644);
    if(fd == -1)
	die("open output file failed: %s\n", strerror(errno));


    void* h_gsobject = write_chunk_start(fd, FILETYPE_GS_OBJECT);
    struct gs_chunk ch;
    struct gs_object obj;
    memset(&obj, 0, sizeof(obj));
    strncpy(obj.name, obj_name, 32);
    obj.quat.w = 1.f;
    
    write(fd, &obj, sizeof(obj));

    void* h_gsobjectdata = write_chunk_start(fd, CHUNK_OBJECT_DATA);

    ch.type = CHUNK_VERTEX_DATA;
    ch.size = vdata_len;
    write(fd, &ch, sizeof(ch));

    uint32_t num = nventries;
    write(fd, &num, sizeof(num));

    unsigned int i;
    struct v_entry *v = v_entries;
    struct gs_vertex *output = malloc(nventries * sizeof(struct gs_vertex));
    struct gs_vertex *vp = output;
    for(i = 0; i < nventries; i++) {
        memcpy(&vp->pos, v_list + v->v_idx*3 , sizeof(vp->pos));
        vp->uv[0] = vt_list[v->vt_idx*2];
        vp->uv[1] = -vt_list[v->vt_idx*2+1];
        memcpy(&vp->nor, vn_list + v->vn_idx*3, sizeof(vp->nor));
        vp->color = 0x80808080;
        v->idx = i;
        v = v->next;
        vp++;
    }
    write(fd, output, sizeof(struct gs_vertex)*nventries);

    ch.type = CHUNK_INDEX_DATA;
    ch.size = idata_len;
    write(fd, &ch, sizeof(ch));

    num = nindices * 3;
    write(fd, &num, sizeof(num));

    struct v_entry **vi = v_indices;
    for(i = 0; i < nindices; i++) {
        uint32_t idx;
        idx = vi[0]->idx;
        write(fd, &idx, sizeof(idx));
        idx = vi[1]->idx;
        write(fd, &idx, sizeof(idx));
        idx = vi[2]->idx;
        write(fd, &idx, sizeof(idx));
        vi += 3;
    }

    ch.type = CHUNK_MATERIAL_DATA;
    ch.size = mdata_len;
    write(fd, &ch, sizeof(ch));

    num = nmaterials;
    write(fd, &num, sizeof(num));

    struct material *m = material_list;
    while(m) {
        ch.type = CHUNK_MATERIAL_DEF;
        ch.size = sizeof(struct gs_material);
        write(fd, &ch, sizeof(ch));
        
        write(fd, m->gsmat, sizeof(*m->gsmat));
        m = m->next;
    }
    if (write_chunk_end(h_gsobjectdata) != objdata_len) {
        die("Internal Chunk Writer error, please report");
    }
    
    // Write gs_path entries
    struct path_data_ll* p = path_list; 
    while(p) {
        ch.type = CHUNK_OBJ_PATH_INFO;
        ch.size = sizeof(p->p);
        write(fd, &ch, sizeof(ch));
        write(fd, &p->p, sizeof(p->p));
        p = p->next;
    }
    
    
    write_chunk_end(h_gsobject);
    close(fd);
}





static void write_material_files()
{
    struct gs_chunk ch;
    char *filename;
    struct material *m = material_list;

    void *statechunkmatdef;
    int fd;

    
    while(m) {
        
        filename = make_path(NULL, m->gsmat->name, ".material_new");
    
    
        printf("writing material: '%s'\n", filename);
        if(!force_overwrite && file_exists(filename))
                die("'%s' already exists. Remove or use -f to force overwrite\n", filename);


        fd = open(filename, CREAT_FLAGS, 0644);
        if(fd == -1)
            die("open output file failed: %s\n", strerror(errno));

        statechunkmatdef = write_chunk_start(fd, CHUNK_MATERIAL_DEF);
        
        write(fd, m->gsmat, sizeof(*m->gsmat));

        if (m->diffusemap) {
            ch.type = CHUNK_MAT_TEXTURE;
            ch.size = sizeof(struct gs_tex);
            write(fd, &ch, sizeof(ch));
            write(fd, m->diffusemap, sizeof(struct gs_tex));
        }

        if (m->specularmap) {
            ch.type = CHUNK_MAT_SPECULARMAP;
            ch.size = sizeof(struct gs_tex);
            write(fd, &ch, sizeof(ch));
            write(fd, m->specularmap, sizeof(struct gs_tex));
        }

        if (m->normalmap) {
            ch.type = CHUNK_MAT_NORMALMAP;
            ch.size = sizeof(struct gs_tex);
            write(fd, &ch, sizeof(ch));
            write(fd, m->normalmap, sizeof(struct gs_tex));
        }
        write_chunk_end(statechunkmatdef);
        close(fd);

        m = m->next;
    }
}

void usage(char *name) {
    printf("Usage: %s <options> <OBJ file>\n"
    " -f force overwrite\n"
    " -v verbose output\n", name);
}

int main(int argc, char **argv)
{
    int i;
    char *filename = NULL, *basefilename = NULL, *path = NULL;

    printf(TOOLNAME "\n");
    if(argc < 2) {
	usage(argv[0]);
	return -1;
    }
   /* if(argc == 3) {
	char *arg = argv[1];
	if(arg[0] == '-' && arg[1] == 'f')
	    force_overwrite = 1;
	argv++;
    } */
    
    for(i=1; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
            if (arg[1] == 'f') {
                force_overwrite = 1;
                
            } else if (arg[1] == 'v') {
                verbose = 1;
                if (arg[2] == '2') {
                    verbose = 2;
                }
            } else {
                printf("Unknown option %s \n", argv[i]);
                usage(argv[0]);
                return -1;
            }
	} else {
            
            filename = strrchr(argv[i], PATH_DELIMITER);
            if(filename) {
	        *filename++ = '\0';
                path = argv[i];
                if(chdir(path) != 0)
                   die("chdir('%s') failed: %s\n", path, strerror(errno));
            } else {
                path = ".";
                filename = argv[i];
            }
        }
    }

    
    if (!filename) {
        usage(argv[0]);
        return -1;
    }
    v_list  = malloc(sizeof(float) * 8 * MAX_LIST_SIZE);
    vn_list = v_list + 3 * MAX_LIST_SIZE;
    vt_list = vn_list + 3 * MAX_LIST_SIZE;

    v_indices = malloc(sizeof(struct v_entry*)*3*MAX_LIST_SIZE);
    
    
    
    basefilename = filename_noext(filename);
    
    printf("working dir: %s\n", path);
    
    read_obj_file(filename);
    char *path_filename = make_path(NULL, basefilename, ".paths");
    char *object_filename = make_path(NULL, basefilename, ".object");
    path_list = read_paths_file(path_filename);

    if(!obj_name) {
	//die("no object name specified\n");
        obj_name = strdup(basefilename);
        print(0, "[warn] No object name found, falling back to %s\n", obj_name);
    }    
    /*printf("object: %s read: %u/%u/%u  indices: %u created v_entries: %u\n",
	   obj_name, num_v, num_vn, num_vt, nindices,
	   nventries);*/

    // FIXME: use name derived from obj/material name
    write_object_file(object_filename);

    write_material_files();
    return 0;
}
